
/**
 * @file ohbbhlp.c
 *
 * See ohbbhlp.h for the detail explanation.
 *
 * This source code can be distributed freely under BSD license.
 *
 * Taeho Oh ( ohhara@postech.edu , http://ohhara.sarang.net , 2003/11/13 )
 */

#include <windows.h>

#define OHBEXPORT __declspec(dllexport)
#include "ohbbhlp.h"

#define OHBINITLOCK() InitializeCriticalSection(&CriticalSection)
#define OHBLOCK() EnterCriticalSection(&CriticalSection)
#define OHBUNLOCK() LeaveCriticalSection(&CriticalSection)
#define OHBDESTROYLOCK() DeleteCriticalSection(&CriticalSection)

#define OHB_TRANS_WIN_CLS "TransWindow"
#define OHB_HOME_DIR_LEN 2048
#define OHB_JK_TRANS_MSG 5393
#define OHB_KJ_TRANS_MSG 5409

typedef struct _OHBTRANSI
{
  LPCWSTR lpStr;
  int fd;
  char reserved[24];
} OHBTRANSI;

typedef struct _OHBABEL
{
  CHAR szJKHomeDir[OHB_HOME_DIR_LEN];
  CHAR szKJHomeDir[OHB_HOME_DIR_LEN];
  HMODULE hJKMod;
  HMODULE hKJMod;
  int (*InitJKTE)(HWND hwnd, const char *homeDir, int data0);
  int (*SetJKTESource)(int fd, const char *srcStr);
  int (*SetJKTEOption)(int fd, int p1, int p2, int p3, int p4, int p5,
    int p6, int p7, int p8, int p9, int p10, int p11, int p12, int p13,
    int p14, const char *srcStr);
  int (*TransJKTE)(int fd, int data0);
  const char *(*GetJKTESolution)(int fd);
  int (*ReleaseJKTESolution)(int fd);
  int (*EndJKTE)(int fd, int fd2);
  int (*InitKJTE)(HWND hwnd, const char *homeDir, int data0);
  int (*SetKJTESource)(int fd, const char *srcStr);
  int (*SetKJTEOption)(int fd, int p1, int p2, int p3, int p4, int p5,
    const char *srcStr);
  int (*TransKJTE)(int fd, int data0);
  const char *(*GetKJTESolution)(int fd);
  int (*ReleaseKJTESolution)(int fd);
  int (*EndKJTE)(int fd, int fd2);
} OHBABEL;

static OHBABEL babel;
static CRITICAL_SECTION CriticalSection;
static HWND hwndTransWindow;

static int OhbInit(void);
static int OhbDestroy(void);
static int OhbCreateTransWindow(void);
static int OhbDestroyTransWindow(void);
static int OhbInitHomeDir(void);
static int OhbDestroyHomeDir(void);
static int OhbInitBabel(void);
static int OhbDestroyBabel(void);
static LRESULT CALLBACK OhbTransWindowMsgProc(HWND hwnd, UINT message,
  WPARAM wParam, LPARAM lParam);

static int OhbInit(void)
{
  OHBINITLOCK();
  if (OhbInitHomeDir() != 0)
  {
    return -1;
  }
  if (OhbInitBabel() != 0)
  {
    return -2;
  }
  if (OhbCreateTransWindow() != 0)
  {
    return -3;
  }

  return 0;
}

static int OhbDestroy(void)
{
  OhbDestroyTransWindow();
  OhbDestroyBabel();
  OhbDestroyHomeDir();
  OHBDESTROYLOCK();

  return 0;
}

static int OhbCreateTransWindow(void)
{
  WNDCLASS wndclass;

  wndclass.style = 0;
  wndclass.lpfnWndProc = OhbTransWindowMsgProc;
  wndclass.cbClsExtra = 0;
  wndclass.cbWndExtra = 0;
  wndclass.hInstance = 0;
  wndclass.hIcon = 0;
  wndclass.hCursor = 0;
  wndclass.hbrBackground = 0;
  wndclass.lpszMenuName = 0;
  wndclass.lpszClassName = TEXT(OHB_TRANS_WIN_CLS);

  RegisterClass(&wndclass);

  hwndTransWindow = CreateWindow(TEXT(OHB_TRANS_WIN_CLS), TEXT("TransApp"), 0,
    0, 0, 0, 0, 0, 0, 0, 0);

  if (hwndTransWindow)
  {
    return 0;
  }

  return -1;
}

static int OhbDestroyTransWindow(void)
{
  if (hwndTransWindow)
  {
    PostMessage(hwndTransWindow, WM_QUIT, 0, 0);
    UnregisterClass(TEXT(OHB_TRANS_WIN_CLS), 0);
  }

  return 0;
}

static int OhbInitBabel(void)
{
  CHAR szDllBuff[OHB_HOME_DIR_LEN];

  lstrcpyA(szDllBuff, babel.szJKHomeDir);
  lstrcatA(szDllBuff, "..\\jkte40.dll");
  babel.hJKMod = LoadLibraryA(szDllBuff);
  if (babel.hJKMod == NULL)
  {
    return -1;
  }

  lstrcpyA(szDllBuff, babel.szKJHomeDir);
  lstrcatA(szDllBuff, "..\\kjtedll.dll");
  babel.hKJMod = LoadLibraryA(szDllBuff);
  if (babel.hKJMod == NULL)
  {
    return -2;
  }

  babel.InitJKTE = (PVOID)GetProcAddress(babel.hJKMod, "InitJKTE");
  babel.SetJKTESource = (PVOID)GetProcAddress(babel.hJKMod, "SetJKTESource");
  babel.SetJKTEOption = (PVOID)GetProcAddress(babel.hJKMod, "SetJKTEOption");
  babel.TransJKTE= (PVOID)GetProcAddress(babel.hJKMod, "TransJKTE");
  babel.GetJKTESolution = (PVOID)GetProcAddress(babel.hJKMod,
    "GetJKTESolution");
  babel.ReleaseJKTESolution = (PVOID)GetProcAddress(babel.hJKMod,
    "ReleaseJKTESolution");
  babel.EndJKTE= (PVOID)GetProcAddress(babel.hJKMod, "EndJKTE");

  babel.InitKJTE = (PVOID)GetProcAddress(babel.hKJMod, "InitKJTE");
  babel.SetKJTESource = (PVOID)GetProcAddress(babel.hKJMod, "SetKJTESource");
  babel.SetKJTEOption = (PVOID)GetProcAddress(babel.hKJMod, "SetKJTEOption");
  babel.TransKJTE= (PVOID)GetProcAddress(babel.hKJMod, "TransKJTE");
  babel.GetKJTESolution = (PVOID)GetProcAddress(babel.hKJMod,
    "GetKJTESolution");
  babel.ReleaseKJTESolution = (PVOID)GetProcAddress(babel.hKJMod,
    "ReleaseKJTESolution");
  babel.EndKJTE= (PVOID)GetProcAddress(babel.hKJMod, "EndKJTE");

  if (babel.InitJKTE == NULL || babel.SetJKTESource == NULL ||
    babel.SetJKTEOption == NULL || babel.TransJKTE == NULL ||
    babel.GetJKTESolution == NULL || babel.ReleaseJKTESolution == NULL ||
    babel.EndJKTE == NULL || babel.InitKJTE == NULL ||
    babel.SetKJTESource == NULL || babel.SetKJTEOption == NULL ||
    babel.TransKJTE == NULL || babel.GetKJTESolution == NULL ||
    babel.ReleaseKJTESolution == NULL || babel.EndKJTE == NULL)
  {
    return -3;
  }

  return 0;
}

static int OhbDestroyBabel(void)
{
  FreeLibrary(babel.hJKMod);
  FreeLibrary(babel.hKJMod);
  ZeroMemory(&babel, sizeof(babel));

  return 0;
}

static int OhbInitHomeDir(void)
{
  HKEY hKey;
  int len, i;
  DWORD dwLen = OHB_HOME_DIR_LEN;

  if (RegOpenKeyExA(HKEY_LOCAL_MACHINE,
    "SOFTWARE\\Unisoft\\Babel TOP 2002\\JSender", 0, KEY_QUERY_VALUE, &hKey)
    != ERROR_SUCCESS)
  {
    return -1;
  }

  if (RegQueryValueExA(hKey, NULL, NULL, NULL, (LPBYTE)babel.szJKHomeDir,
    &dwLen) != ERROR_SUCCESS)
  {
    return -2;
  }

  RegCloseKey(hKey);

  babel.szJKHomeDir[OHB_HOME_DIR_LEN - 1] = '\0';

  len = lstrlenA(babel.szJKHomeDir);
  if (len > 2000)
  {
    return -3;
  }

  for (i = len - 1; i >= 0; i--)
  {
    if (babel.szJKHomeDir[i] == '\\')
    {
      babel.szJKHomeDir[i] = '\0';
      lstrcatA(babel.szJKHomeDir, "\\dicJK\\");
      lstrcpy(babel.szKJHomeDir, babel.szJKHomeDir);
      babel.szKJHomeDir[i] = '\0';
      lstrcatA(babel.szKJHomeDir, "\\dicKJ\\");
      break;
    }
  }

  if (i < 0)
  {
    return -4;
  }

  return 0;
}

static int OhbDestroyHomeDir(void)
{
  babel.szJKHomeDir[0] = '\0';
  babel.szKJHomeDir[0] = '\0';

  return 0;
}

static LRESULT CALLBACK OhbTransWindowMsgProc(HWND hwnd, UINT message,
  WPARAM wParam, LPARAM lParam)
{
  if (hwnd != 0 && (message == OHB_JK_TRANS_MSG ||
    message == OHB_KJ_TRANS_MSG))
  {
    PostMessage(hwnd, message, wParam, lParam);
  }

  return DefWindowProc(hwnd, message, wParam, lParam);
}

BOOL WINAPI DllMain(HINSTANCE hinstDll, DWORD fdwReason, PVOID fImpLoad)
{
  switch (fdwReason)
  {
    case DLL_PROCESS_ATTACH:
      if (OhbInit() != 0)
      {
        OhbDestroy();
        return FALSE;
      }
      break;

    case DLL_PROCESS_DETACH:
      OhbDestroy();
      break;
  }

   return TRUE;
}

OHBEXPORT OHBTRANS * OHBCALL OhbJKTrans(LPCWSTR jpStr, PVOID reserved)
{
  OHBTRANSI *ohbTrans;
  LPSTR szSrc;
  LPWSTR szDst;
  LPCSTR szTrans;
  int len;
  MSG msg;
  int fd;

  OHBLOCK();

  len = WideCharToMultiByte(932, 0, jpStr, -1, 0, 0, 0, 0);
  szSrc = HeapAlloc(GetProcessHeap(), 0, len);
  WideCharToMultiByte(932, 0, jpStr, -1, szSrc, len, 0, 0);

  fd = babel.InitJKTE(hwndTransWindow, babel.szJKHomeDir, 0);
  babel.SetJKTESource(fd, szSrc);
  babel.SetJKTEOption(fd, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, szSrc);
  babel.TransJKTE(fd, 0);
  GetMessage(&msg, hwndTransWindow, 0, 0);
  szTrans = babel.GetJKTESolution(fd);

  len = MultiByteToWideChar(949, 0, szTrans, -1, 0, 0);
  szDst = HeapAlloc(GetProcessHeap(), 0, len * 2);
  MultiByteToWideChar(949, 0, szTrans, -1, szDst, len);

  ohbTrans = HeapAlloc(GetProcessHeap(), 0, sizeof(OHBTRANSI));
  ohbTrans->fd = fd;
  ohbTrans->lpStr = szDst;
  HeapFree(GetProcessHeap(), 0, szSrc);

  OHBUNLOCK();

  return (OHBTRANS *)ohbTrans;
}

OHBEXPORT void OHBCALL OhbJKFree(OHBTRANS *ohbTrans)
{
  OHBTRANSI *ohbTransi;
  int fd;

  OHBLOCK();
  ohbTransi = (OHBTRANSI *)ohbTrans;
  fd = ohbTransi->fd;
  babel.ReleaseJKTESolution(fd);
  babel.EndJKTE(fd, fd);
  HeapFree(GetProcessHeap(), 0, (PVOID)ohbTrans->lpStr);
  HeapFree(GetProcessHeap(), 0, ohbTrans);
  OHBUNLOCK();
}

OHBEXPORT OHBTRANS * OHBCALL OhbKJTrans(LPCWSTR krStr, PVOID reserved)
{
  OHBTRANSI *ohbTrans;
  LPSTR szSrc;
  LPWSTR szDst;
  LPCSTR szTrans;
  int len;
  MSG msg;
  int fd;

  OHBLOCK();

  len = WideCharToMultiByte(949, 0, krStr, -1, 0, 0, 0, 0);
  szSrc = HeapAlloc(GetProcessHeap(), 0, len);
  WideCharToMultiByte(949, 0, krStr, -1, szSrc, len, 0, 0);

  fd = babel.InitKJTE(hwndTransWindow, babel.szKJHomeDir, 0);
  babel.SetKJTESource(fd, szSrc);
  babel.SetKJTEOption(fd, 0, 0, 0, 0, 0, szSrc);
  babel.TransKJTE(fd, 0);
  GetMessage(&msg, hwndTransWindow, 0, 0);
  szTrans = babel.GetKJTESolution(fd);

  len = MultiByteToWideChar(20932, 0, szTrans, -1, 0, 0);
  szDst = HeapAlloc(GetProcessHeap(), 0, len * 2);
  MultiByteToWideChar(20932, 0, szTrans, -1, szDst, len);

  ohbTrans = HeapAlloc(GetProcessHeap(), 0, sizeof(OHBTRANSI));
  ohbTrans->fd = fd;
  ohbTrans->lpStr = szDst;
  HeapFree(GetProcessHeap(), 0, szSrc);

  OHBUNLOCK();

  return (OHBTRANS *)ohbTrans;
}

OHBEXPORT void OHBCALL OhbKJFree(OHBTRANS *ohbTrans)
{
  OHBTRANSI *ohbTransi;
  int fd;

  OHBLOCK();
  ohbTransi = (OHBTRANSI *)ohbTrans;
  fd = ohbTransi->fd;
  babel.ReleaseKJTESolution(fd);
  babel.EndKJTE(fd, fd);
  HeapFree(GetProcessHeap(), 0, (PVOID)ohbTrans->lpStr);
  HeapFree(GetProcessHeap(), 0, ohbTrans);
  OHBUNLOCK();
}
